#ifdef _WINDOWS
#define WINDOWS
#endif

#include "ac_plugin.h"
#include "Undoables.h"
#include <stdio.h>
#include <string.h>

//---------------------------------------------------------------------------------------
class CSelConnItem
{
public:
   void *address;
   bool selected;
   CSelConnItem()
   {
      address = NULL;
      selected = false;
   }
};
//---------------------------------------------------------------------------------------
void ddhselconn_sort_address_list( CSelConnItem *arr, int left, int right )
{

   int i = left;
   int j = right;
   void *x = arr[( left + right ) / 2].address;
   CSelConnItem seltemp;

   do
   {
      while( arr[i].address < x && i < right )
      {
         ++i;
      }
      while( arr[j].address > x && j > left )
      {
         --j;
      }
      if( i <= j )
      {
         seltemp.address = arr[i].address;
         seltemp.selected = arr[i].selected;
         arr[i].address = arr[j].address;
         arr[i].selected = arr[j].selected;
         arr[j].address = seltemp.address;
         arr[j].selected = seltemp.selected;
         ++i;
         --j;
      }
   } while( i <= j );

   if( left < j )
   {
      ddhselconn_sort_address_list( arr, left, j );
   }
   if( i < right )
   {
      ddhselconn_sort_address_list( arr, i, right );
   }

}
//---------------------------------------------------------------------------------------
int ddhselconn_search_address_list( CSelConnItem *arr, int count, void *address )
{
   int low = 0;
   int high = count - 1;
   int mid;
   while( low <= high )
   {
      mid = ( low + high ) / 2;
      if( address < arr[mid].address )
      {
         high = mid - 1;
      }
      else if( address > arr[mid].address )
      {
         low = mid + 1;
      }
      else
      {
         return mid;
      }
   }
   return -1;
}
//---------------------------------------------------------------------------------------
static void ddhselconn_go()
{

   // Bail if nothing selected, or we're not in surface/vertex mode
   if( nothing_is_selected() )
   {
      message_dialog( "selconnected : Nothing selected" );
      return;
   }

   int selmode = ac_get_select_mode();
   if( selmode != SELECT_SURFACE && selmode != SELECT_VERTEX )
   {
      message_dialog( "selconnected : Must be in Surface or Vertex mode" );
      return;
   }

   List *selsurfs = NULL;
   List *selobjs = NULL;

   selsurfs = ac_selection_get_whole_surfaces_all();
   if( selsurfs == NULL )
   {
      message_dialog( "selconn : no whole surfaces selected" );
      return;
   }
   for( List *ps = selsurfs; ps != NULL; ps = ps->next )
   {
      list_add_item_unique( &selobjs, object_of_surface(( Surface * )( ps->data )));
   }

   // Grab the maximum number of selected surfaces we can have, based on the current selection
   int numsurfs = 0;
   for( List *po = selobjs; po != NULL; po = po->next )
   {
      numsurfs += ac_object_get_num_surfaces(( ACObject * )( po->data ));
   }

   // Build our address list with the elements of the selected objects
   // Grab the edges of all selected objects
   CSelConnItem *selarr = new CSelConnItem[numsurfs];
   int ctr = 0;
   List *edges = NULL;
   for( List *po = selobjs; po != NULL; po = po->next )
   {
      ACObject *o = ( ACObject * )( po->data );
      for( List *ps = ac_object_get_surfacelist( o ); ps != NULL; ps = ps->next )
      {
         selarr[ctr++].address = ps->data;
      }
      list_append_list( &edges, ac_object_get_edges( o ));
   }
   list_free( &selobjs );

   // Sort the address list
   ddhselconn_sort_address_list( selarr, 0, numsurfs - 1 );

   // Select all conected surfaces
   for( List *ps = selsurfs; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      int idx = ddhselconn_search_address_list( selarr, numsurfs, s );
      if( selarr[idx].selected )
      {
         // Already grabbed this surface
         continue;
      }
      selarr[idx].selected = true;
      // Get edges
      List *el = ac_edgelist_get_by_surface( edges, s );
      for( List *pe = el; pe != NULL; pe = pe->next )
      {
         ACEdge *e = ( ACEdge * )( pe->data );
         for( List *pes = e->surfacelist; pes != NULL; pes = pes->next )
         {
            Surface *s2 = ( Surface * )( pes->data );
            int sidx = ddhselconn_search_address_list( selarr, numsurfs, s2 );
            if( !( selarr[sidx].selected ))
            {
               list_add_item( &ps, s2 );
            }
         }
      }
   }

   // Cleanup
   ac_edgelist_free( &edges );
   list_free( &selsurfs );

   // Create a list of selected surfaces
   selsurfs = NULL;
   ctr = 0;
   for( int ix = 0; ix < numsurfs; ++ix )
   {
      if( selarr[ix].selected )
      {
         list_add_item_head( &selsurfs, selarr[ix].address );
         ++ctr;
      }
   }
   display_message( "Selected %d surfaces", ctr );
   add_undoable_change_selection( "select connected surfaces" );
   ac_selection_select_surfacelist( selsurfs );

   // Cleanup
   delete[] selarr;
   list_free( &selsurfs );

   redraw_all();
   display_status();

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{

   static bool firstTime = true;
   if( firstTime )
   {

      ac_add_command( "ddhselconn_go", ddhselconn_go );
      ac_add_tools_menu_item( "Select connected surfaces", "ac3d ddhselconn_go",
                              "Select surfaces connected to the currently selected surfaces by shared edges" );

      firstTime = false;

   }

   return 0;

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Select Connected Surfaces - version 1.0 - Dennis Hawthorne" );
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Select Connected Surfaces - Dennis Hawthorne - Select surfaces connected to the currently selected surfaces by shared edges." );
}
//---------------------------------------------------------------------------------------
